/*  Copyright (C) 2017 Nicholas C. Strauss (strauss@positive-internet.com)
    This file is part of ORBTIMES.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <math.h>
#include "expression.h"
#include "evaluate.h"
#include "location_expression.h"
#include "parser.h"
#include "lexer.h"
#include <stdio.h>
#include <search.h>
#include "main.h"
#define MAIN extern
#include "globals.h"

struct hsearch_data   *mactab;
funcall_t funcalls[i_maxFuncalls] = {
    {  i_nighttime          , "is_nighttime", handle_isnighttime, },
    {  i_daytime            , "is_daytime", handle_isdaytime, },  
    
    {  i_sunrise            , "sunrise", handle_sunrise, },
    {  i_sunset             , "sunset",  handle_sunset, },
    {  i_azimuth            , "azimuth",  handle_azimuth, },
    {  i_altitude           , "altitude",  handle_altitude, },
    //moon
    {  i_moonrise            , "moonrise", handle_moonrise, },
    {  i_moonset             , "moonset",  handle_moonset, },
    {  i_moon_azimuth            , "moon_azimuth",  handle_moon_azimuth, },
    {  i_moon_altitude           , "moon_altitude",  handle_moon_altitude, },
    {  i_moon_nighttime          , "is_moon_nighttime", handle_moon_isnighttime, },
    {  i_moon_daytime            , "is_moon_daytime", handle_moon_isdaytime, },  
    //mercury
    {  i_mercuryrise            , "mercuryrise", handle_mercuryrise, },
    {  i_mercuryset             , "mercuryset",  handle_mercuryset, },
    {  i_mercury_azimuth            , "mercury_azimuth",  handle_mercury_azimuth, },
    {  i_mercury_altitude           , "mercury_altitude",  handle_mercury_altitude, },
    {  i_mercury_nighttime          , "is_mercury_nighttime", handle_mercury_isnighttime, },
    {  i_mercury_daytime            , "is_mercury_daytime", handle_mercury_isdaytime, },  
    //venus
    {  i_venusrise            , "venusrise", handle_venusrise, },
    {  i_venusset             , "venusset",  handle_venusset, },
    {  i_venus_azimuth            , "venus_azimuth",  handle_venus_azimuth, },
    {  i_venus_altitude           , "venus_altitude",  handle_venus_altitude, },
    {  i_venus_nighttime          , "is_venus_nighttime", handle_venus_isnighttime, },
    {  i_venus_daytime            , "is_venus_daytime", handle_venus_isdaytime, },  
    //mars
    {  i_marsrise            , "marsrise", handle_marsrise, },
    {  i_marsset             , "marsset",  handle_marsset, },
    {  i_mars_azimuth            , "mars_azimuth",  handle_mars_azimuth, },
    {  i_mars_altitude           , "mars_altitude",  handle_mars_altitude, },
    {  i_mars_nighttime          , "is_mars_nighttime", handle_mars_isnighttime, },
    {  i_mars_daytime            , "is_mars_daytime", handle_mars_isdaytime, },  
    //jupiter
    {  i_jupiterrise            , "jupiterrise", handle_jupiterrise, },
    {  i_jupiterset             , "jupiterset",  handle_jupiterset, },
    {  i_jupiter_azimuth            , "jupiter_azimuth",  handle_jupiter_azimuth, },
    {  i_jupiter_altitude           , "jupiter_altitude",  handle_jupiter_altitude, },
    {  i_jupiter_nighttime          , "is_jupiter_nighttime", handle_jupiter_isnighttime, },
    {  i_jupiter_daytime            , "is_jupiter_daytime", handle_jupiter_isdaytime, },  
    //saturn
    {  i_saturnrise            , "saturnrise", handle_saturnrise, },
    {  i_saturnset             , "saturnset",  handle_saturnset, },
    {  i_saturn_azimuth            , "saturn_azimuth",  handle_saturn_azimuth, },
    {  i_saturn_altitude           , "saturn_altitude",  handle_saturn_altitude, },
    {  i_saturn_nighttime          , "is_saturn_nighttime", handle_saturn_isnighttime, },
    {  i_saturn_daytime            , "is_saturn_daytime", handle_saturn_isdaytime, },  
    //uranus
    {  i_uranusrise            , "uranusrise", handle_uranusrise, },
    {  i_uranusset             , "uranusset",  handle_uranusset, },
    {  i_uranus_azimuth            , "uranus_azimuth",  handle_uranus_azimuth, },
    {  i_uranus_altitude           , "uranus_altitude",  handle_uranus_altitude, },
    {  i_uranus_nighttime          , "is_uranus_nighttime", handle_uranus_isnighttime, },
    {  i_uranus_daytime            , "is_uranus_daytime", handle_uranus_isdaytime, },  
    //neptune
    {  i_neptunerise            , "neptunerise", handle_neptunerise, },
    {  i_neptuneset             , "neptuneset",  handle_neptuneset, },
    {  i_neptune_azimuth            , "neptune_azimuth",  handle_neptune_azimuth, },
    {  i_neptune_altitude           , "neptune_altitude",  handle_neptune_altitude, },
    {  i_neptune_nighttime          , "is_neptune_nighttime", handle_neptune_isnighttime, },
    {  i_neptune_daytime            , "is_neptune_daytime", handle_neptune_isdaytime, },  

    {  i_now                , "now",     handle_now, },

    {  i_month              , "month",   handle_month, },
    {  i_dow                , "dow",     handle_dow, },
    {  i_year               , "year",    handle_year, },                

    {  i_jan                , "jan",     handle_jan, },
    {  i_feb                , "feb",     handle_feb, },
    {  i_mar                , "mar",     handle_mar, },
    {  i_apr                , "apr",     handle_apr, },
    {  i_may                , "may",     handle_may, },    
    {  i_jun                , "jun",     handle_jun, },
    {  i_jul                , "jul",     handle_jul, },
    {  i_aug                , "aug",     handle_aug, },
    {  i_sep                , "sep",     handle_sep, },
    {  i_oct                , "oct",     handle_oct, },
    {  i_nov                , "nov",     handle_nov, },
    {  i_dec                , "dec",     handle_dec, },            
};

void
initialize_cvalues()
{
  ENTRY        e, *ep;
  int    i, success;
  mactab=(struct hsearch_data *)calloc(i_maxFuncalls, sizeof(struct hsearch_data));
  success=hcreate_r(i_maxFuncalls, mactab);
  if (!success)
    printf("hcreate_r: ran out of memory %d\n", errno);

  for (i=0;i<i_maxFuncalls;i++){
    e.key = funcalls[i].str;
    e.data = (void*)&funcalls[i];
    success = hsearch_r(e, ENTER, &ep, mactab);  
    if (!success)
      printf("hcreate_r: ran out of memory %d\n", errno);
  }
}

double
evaluate_cvalue(char *cvalue, bool *truth)
{
  int       success;
  ENTRY     e, *ep;
  *truth = 0;
  e.key = cvalue;
  e.data = (void*)NULL;
  success = hsearch_r(e, FIND, &ep, mactab);  
  if (ep){
    *truth = 1;
    return ((funcall_t*) (ep->data))->handler();
  }
  else return 0;
}

/* 0.0 is false, 
   non-zero is true */
double 
evaluate(SExpression *e)
{
  if (!e) return 0.0;             // handle unary minus/plus

  switch (e->type) {
  case eVALUE:
    return e->value;
  case eVARIABLE:
    {
      bool   truth;
      double value = evaluate_cvalue(e->cvalue, &truth);
      printf("evaluate: cvalue %s value %lf %d\n", e->cvalue, value, truth);
      return value;
    }
  case eLESS:
    {
      double
	l = evaluate(e->left),
	m = evaluate(e->right);
      if (isnan(l) ||  isnan(m)) return nan("false");
      if (l <= m) return m;
      else return nan("false");
    }
  case eMORE:
    {
      double
	l = evaluate(e->left),
	m = evaluate(e->right);
      if (isnan(l) ||  isnan(m)) return nan("false");
      if (l >= m) return m;
      else return nan("false");
    }
  case eAND:
    {
      double
	l = evaluate(e->left),
	m = evaluate(e->right);
      if (isnan(l) ||  isnan(m)) return nan("false");
      return l && m;
    }
  case eOR:
    {    
      double
	l = evaluate(e->left),
	m = evaluate(e->right);
      if (isnan(l) ||  isnan(m)) return nan("false");
      return l || m;
    }
  case ePLUS:
    {
    return evaluate(e->left) + evaluate(e->right);
    }
  case eMINUS:
    {
    return evaluate(e->left) - evaluate(e->right);
    }
  case eTIMES:
    {
    return evaluate(e->left) * evaluate(e->right);
    }
  case eDIVISION:
    {
    return evaluate(e->left) / evaluate(e->right);
    }
  default:
    // shouldn't be here
    return nan("error");
  }
}

double handle_isnighttime()
{
  return nighttime;
}

double handle_isdaytime()
{
  return daytime;
}

double handle_sunrise()
{
  return jdays_sunrise;
}

double handle_sunset()
{
  return jdays_sunset;
}

double handle_azimuth()
{
  return sun_azimuth;
}

double handle_altitude()
{
  return sun_altitude;
}
// Moon
double handle_moonrise()
{
  return jdays_moonrise;
}

double handle_moonset()
{
  return jdays_moonset;
}

double handle_moon_isnighttime()
{
  return moon_nighttime;
}

double handle_moon_isdaytime()
{
  return moon_daytime;
}

double handle_moon_azimuth()
{
  return jdays_moon_azimuth;
}

double handle_moon_altitude()
{
  return jdays_moon_altitude;
}

// Mercury
double handle_mercuryrise()
{
  return jdays_planets_rise[MERCURY];
}

double handle_mercuryset()
{
  return jdays_planets_set[MERCURY];
}

double handle_mercury_azimuth_rise()
{
  return jdays_planets_azimuth_rise[MERCURY];
}

double handle_mercury_azimuth_set()
{
  return jdays_planets_azimuth_set[MERCURY];
}

double handle_mercury_azimuth()
{
  return jdays_planets_azimuth[MERCURY];
}

double handle_mercury_altitude()
{
  return jdays_planets_alitude[MERCURY];
}

double handle_mercury_isnighttime()
{
  return planets_nighttime[MERCURY];
}

double handle_mercury_isdaytime()
{
  return planets_daytime[MERCURY];
}

// Venus
double handle_venusrise()
{
  return jdays_planets_rise[VENUS];
}

double handle_venusset()
{
  return jdays_planets_set[VENUS];
}

double handle_venus_azimuth_rise()
{
  return jdays_planets_azimuth_rise[VENUS];
}

double handle_venus_azimuth_set()
{
  return jdays_planets_azimuth_set[VENUS];
}

double handle_venus_azimuth()
{
  return jdays_planets_azimuth[VENUS];
}

double handle_venus_altitude()
{
  return jdays_planets_alitude[VENUS];
}

double handle_venus_isnighttime()
{
  return planets_nighttime[VENUS];
}

double handle_venus_isdaytime()
{
  return planets_daytime[VENUS];
}
// Mars
double handle_marsrise()
{
  return jdays_planets_rise[MARS];
}

double handle_marsset()
{
  return jdays_planets_set[MARS];
}

double handle_mars_azimuth_rise()
{
  return jdays_planets_azimuth_rise[MARS];
}

double handle_mars_azimuth_set()
{
  return jdays_planets_azimuth_set[MARS];
}

double handle_mars_azimuth()
{
  return jdays_planets_azimuth[MARS];
}

double handle_mars_altitude()
{
  return jdays_planets_alitude[MARS];
}

double handle_mars_isnighttime()
{
  return planets_nighttime[MARS];
}

double handle_mars_isdaytime()
{
  return planets_daytime[MARS];
}
// Jupiter
double handle_jupiterrise()
{
  return jdays_planets_rise[JUPITER];
}

double handle_jupiterset()
{
  return jdays_planets_set[JUPITER];
}

double handle_jupiter_azimuth_rise()
{
  return jdays_planets_azimuth_rise[JUPITER];
}

double handle_jupiter_azimuth_set()
{
  return jdays_planets_azimuth_set[JUPITER];
}

double handle_jupiter_azimuth()
{
  return jdays_planets_azimuth[JUPITER];
}

double handle_jupiter_altitude()
{
  return jdays_planets_alitude[JUPITER];
}

double handle_jupiter_isnighttime()
{
  return planets_nighttime[JUPITER];
}

double handle_jupiter_isdaytime()
{
  return planets_daytime[JUPITER];
}
// Saturn
double handle_saturnrise()
{
  return jdays_planets_rise[SATURN];
}

double handle_saturnset()
{
  return jdays_planets_set[SATURN];
}

double handle_saturn_azimuth_rise()
{
  return jdays_planets_azimuth_rise[SATURN];
}

double handle_saturn_azimuth_set()
{
  return jdays_planets_azimuth_set[SATURN];
}

double handle_saturn_azimuth()
{
  return jdays_planets_azimuth[SATURN];
}

double handle_saturn_altitude()
{
  return jdays_planets_alitude[SATURN];
}

double handle_saturn_isnighttime()
{
  return planets_nighttime[SATURN];
}

double handle_saturn_isdaytime()
{
  return planets_daytime[SATURN];
}
// Uranus
double handle_uranusrise()
{
  return jdays_planets_rise[URANUS];
}

double handle_uranusset()
{
  return jdays_planets_set[URANUS];
}

double handle_uranus_azimuth_rise()
{
  return jdays_planets_azimuth_rise[URANUS];
}

double handle_uranus_azimuth_set()
{
  return jdays_planets_azimuth_set[URANUS];
}

double handle_uranus_azimuth()
{
  return jdays_planets_azimuth[URANUS];
}

double handle_uranus_altitude()
{
  return jdays_planets_alitude[URANUS];
}

double handle_uranus_isnighttime()
{
  return planets_nighttime[URANUS];
}

double handle_uranus_isdaytime()
{
  return planets_daytime[URANUS];
}
// Neptune
double handle_neptunerise()
{
  return jdays_planets_rise[NEPTUNE];
}

double handle_neptuneset()
{
  return jdays_planets_set[NEPTUNE];
}

double handle_neptune_azimuth_rise()
{
  return jdays_planets_azimuth_rise[NEPTUNE];
}

double handle_neptune_azimuth_set()
{
  return jdays_planets_azimuth_set[NEPTUNE];
}

double handle_neptune_azimuth()
{
  return jdays_planets_azimuth[NEPTUNE];
}

double handle_neptune_altitude()
{
  return jdays_planets_alitude[NEPTUNE];
}

double handle_neptune_isnighttime()
{
  return planets_nighttime[NEPTUNE];
}

double handle_neptune_isdaytime()
{
  return planets_daytime[NEPTUNE];
}


double handle_now()
{
  return jdays_now;
}

double handle_month()
{
  return now_month;
}

double handle_dow()
{
  return now_dayOfWeek;
}

double handle_year()
{
  return now_year;
}

/* double handle_season() */
/* { */
/*   now_month */
/* } */

double handle_jan(){  return 1; }
double handle_feb(){  return 2; }
double handle_mar(){  return 3; }
double handle_apr(){  return 4; }
double handle_may(){  return 5; }
double handle_jun(){  return 6; }
double handle_jul(){  return 7; }
double handle_aug(){  return 8; }
double handle_sep(){  return 9; }
double handle_oct(){  return 10; }
double handle_nov(){  return 11; }
double handle_dec(){  return 12; }

