/*  Copyright (C) 2017 Nicholas C. Strauss (strauss@positive-internet.com)
    This file is part of ORBTIMES.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

  Expression.h
  Definition of the structure used to build the syntax tree.
 */
#ifndef __EXPRESSION_H__
#define __EXPRESSION_H__

typedef enum tagEOperationType
  {
    eVALUE,
    eVARIABLE,    
    eLESS,
    eMORE,
    eAND,
    eOR,
    ePLUS,
    eMINUS,
    eTIMES,
    eDIVISION,
  } EOperationType;

typedef struct tagSExpression
{
  EOperationType type;///< type of operation

  bool      truth;
  char     *cvalue;
  double    value;
  struct tagSExpression *left; ///< left side of the tree
  struct tagSExpression *middle; // middle tree (not used often)
  struct tagSExpression *right;///< right side of the tree
} SExpression;


SExpression *createNumber(double value);
SExpression *createVariable(char *cvalue);
SExpression *createOperation(EOperationType type, SExpression *left, SExpression *right);
SExpression *createUnaryOperation(EOperationType type, SExpression *right);
SExpression *createTriamory(EOperationType type, SExpression *left, SExpression *middle, SExpression *right);
void deleteExpression(SExpression *b);

#endif // __EXPRESSION_H__
