/*  Copyright (C) 2017 Nicholas C. Strauss (strauss@positive-internet.com)
    This file is part of ORBTIMES.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <math.h>
#include <stdio.h>
#include <strings.h>
#include "startime.h"
#include "sun.h"
#include "moon2.h"
#include "moon3.h"
#include "equationOfTime.h"
#define MAIN extern
#include "globals.h"

#define sind(x) sin(M_PI/180*x)
#define cosd(x) cos(M_PI/180*x)
#define atand(x) 180/M_PI*atan(x)

void compute_moon_az_el(extended dj, extended ut, extended tz, extended ds,
			extended gl, extended gp, extended solar_lon,
			extended &ecliptic_lat, extended &ecliptic_lon,
			extended &ra, extended &dec, extended &value_min, extended &value_max, extended &az, extended &el, extended &phase, extended &distance_to_moon)
{
  double         lha, horizontal_parallax, di,   // calculated parallax at altitude
    app_lha, app_dec, hour_angle,
    solar_longitude, solar_radius; 
  double         ht = 0.0; // assume sea level elevation

  calculate_moon(dj, ut, ecliptic_lon, ecliptic_lat, horizontal_parallax, distance_to_moon, phase, solar_longitude, solar_radius, true);
  //  printf("solar_lon %lf %lf\n", solar_longitude, solar_lon);
  horizontal_parallax *= M_PI/180.;
  di = 2.7249e-1 * sin(horizontal_parallax) + 9.8902e-3 - horizontal_parallax;  // parallax
  geocentric_ecliptic_latlon_to_radec(-1, ra, dec, ecliptic_lon, ecliptic_lat, dj);   // ra in hours
  // add in parallax here for apparent ra, dec
  lha = local_mean_sidereal_time(dj,ut,gl) - ra/15.0;  // ra from degrees to hours
  //  parallax(app_lha, app_dec, gp, ht, lha, dec, horizontal_parallax, false);
  compute_parallax(app_lha, app_dec, lha, dec, gl, gp, dj, ut, ht, horizontal_parallax);
  lha = app_lha;
  dec = app_dec;
  //  printf("lha %lf app_lha %lf dec %lf app_dec %lf\n", lha, app_lha, dec, app_dec);
  observer_coord(lha*15, dec, gp, az, el);
  jdays_moon_azimuth = az;
  jdays_moon_altitude = el;
}

/*
   p apparent hour angle (hours)
   q apparent dec (degrees)
   x true hour angle   (hours)
   y true dec          (degrees)
   gl longitude (deg)
   gp latitude (deg)
   ht elevation
   rp is distance of earth to planet in km.
  hp:earth_radius/rp,
  hp:asin(hp),   (horizontal parallax)
*/
void
compute_parallax(double &p, double &q, double x, double y, double gl, double gp, double dj, double ut, double ht, double hp)
{
  double     
    earth_radius = 6378.16,
    u, rs, rc, rp, dx;

  x *= 15; 
  ht /= 6378140.;
  u = atand(9.96647e-1 * sind(gp) / cosd(gp));
  rs = 9.96647e-1 * sind(u) + ht * sind(gp);
  rc = cosd(u)+ht*cosd(gp);
  rp = 1/sind(hp);
  dx = rc * sind(x) / (rp * cosd(y) - rc * cosd(x));
  dx = atand(dx);
  p = (x + dx);
  q = cosd(p) * (rp*sind(y)-rs) / (rp*cosd(y)*cosd(x)-rc);
  q = atand(q);
  p /= 15.;
}
