/*  Copyright (C) 2017 Nicholas C. Strauss (strauss@positive-internet.com)
    This file is part of ORBTIMES.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <string.h>
#include "parse_options.h"
#include <time.h>   // for timezone
#include "startime.h"
#include <locale.h>
#include "config.h"
#include "date_expression.h"
#include <math.h>
//#include <regex.h>
#include <ctype.h>

extern Options sun_options;
//extern char glocation[255];

int
parse_options (int argc, char **argv)
{
  int c;

  while (1)
    {
            static struct option long_options[] =
	      {
		{"ascii-art", no_argument, 0, 'A'},
		{"check", no_argument, 0, 'c'},
		{"date", required_argument, 0, 'D'},
		{"datejd", required_argument, 0, 'J'},		
		{"daylight", required_argument, 0, 'd'},
		{"twilight_angle", required_argument, 0, 'e'},		
		{"help",    no_argument,       0, 'h'},		
		{"location", required_argument, 0, 'l'},
		{"moon", no_argument, 0, 'M'},				
		{"orbs", no_argument, 0, 'O'},
		{"mercury", no_argument, 0, '1'},
		{"venus", no_argument, 0, '2'},
		{"mars", no_argument, 0, '4'},
		{"jupiter", no_argument, 0, '5'},
		{"saturn", no_argument, 0, '6'},
		{"uranus", no_argument, 0, '7'},
		{"neptune", no_argument, 0, '8'},												
		{"query", required_argument, 0, 'q'},
		{"selftest", no_argument,      0, 's'},
		{"timezone", required_argument, 0, 't'},
		{"twilight", required_argument, 0, 'T'},		
		{"verbose",   no_argument,     0, 'v'},
		{"version",   no_argument,     0, 'V'},
		{0, 0, 0, 0}
	      };
	    /* getopt_long stores the option index here. */
	    int option_index = 0;

	    c = getopt_long (argc, argv, "AcD:d:e:hJ:MOl:q:st:T:vV",
			     long_options, &option_index);

	    /* Detect the end of the options. */
	    if (c == -1)
	      break;

	    switch (c)
	      {

        	// ascii-art
	      case 'A':
		printf ("ASCII-ART -A\n");
		sun_options.self_ascii_art = true;
		break;
		
        	// selftest
	      case 'c':
	      case 's':
		printf ("SELFTEST -s\n");
		sun_options.self_test = true;
		break;

                // verbose
	      case 'v':
		//puts ("option -v\n");
		sun_options.self_verbose = true;
		sun_options.self_verbosity++;
		break;

		// version 
	      case 'V':
		printf("%s\n", PACKAGE_STRING);
		exit(0);
		break;

         	// date_is_now false. input date.
	      case 'D':
		// printf ("option -D with value `%s'\n", optarg);
	        // printf("user specified date_is_now true %s\n", optarg);
		sun_options.self_date_is_now = true;
		abend_date(optarg);
		strcpy(sun_options.date, optarg);
		sun_options.datejd = parse_date(sun_options.date);
		break;
		
         	// daylight savings hour
	      case 'd':
		//		printf ("option -d with value `%s'\n", optarg);
		//		printf("user specified daylight savings override %s\n", optarg);
		sun_options.daylight = atof(optarg);
		sun_options.self_guess_daylight_savings = false;
		break;

         	// date_is_now false. input date. Julian Day.
	      case 'J':
		{
		  double       input_jd;
		// printf ("option -J with value `%s'\n", optarg);
		sun_options.self_date_is_now = true;
		abend_julian(optarg);
		input_jd = atof(optarg)-0.5;
		sun_options.datejd = JulianDays1900(input_jd);
		DMYTimeString(sun_options.date, sun_options.datejd);
		}
		break;

		// timezone hour
	      case 't':
		//		printf ("option -t with value `%s'\n", optarg);
		//		printf("user specified timezone override %s\n", optarg);
		abend_timezone(optarg);
		sun_options.timezone = atof(optarg);
		sun_options.timezone_override = true;
		break;

		// twilight
	      case 'T':
		//		printf ("option -T with value `%s'\n", optarg);
		//		printf("user specified twilight %s\n", optarg);
		int twilight_value;
		sun_options.self_disp_twilight = true;
		abend_twilight(optarg,twilight_value);
		if (twilight_value==0) sun_options.mode_twilight = CIVIL_TWILIGHT;
		else if (twilight_value==1) sun_options.mode_twilight = NAUTICAL_TWILIGHT;
		else if (twilight_value==2) sun_options.mode_twilight = ASTRONOMICAL_TWILIGHT;
		else if (twilight_value==3) sun_options.mode_twilight = GENERAL_TWILIGHT;
		else sun_options.mode_twilight = CIVIL_TWILIGHT;
		break;
		
		// twilightangle
	      case 'e':
		//		printf ("option -T with value `%s'\n", optarg);
		//		printf("user specified twilight %s\n", optarg);
		abend_twilight_angle(optarg);
		sun_options.twilight_angle = atof(optarg);
		break;

        	// Moon
	      case 'M':
		//		printf ("MOON (-M)\n");
		sun_options.self_disp_moon = true;
		break;

        	// Orbs
	      case 'O':
		sun_options.self_disp_mercury = true;
		sun_options.self_disp_venus = true;
		sun_options.self_disp_mars = true;
		sun_options.self_disp_jupiter = true;
		sun_options.self_disp_saturn = true;
		sun_options.self_disp_uranus = true;
		sun_options.self_disp_neptune = true;		
		break;

        	// Mercury
	      case '1':
		sun_options.self_disp_mercury = true;
		break;
        	// Venus
	      case '2':
		sun_options.self_disp_venus = true;
		break;
        	// Mars
	      case '4':
		sun_options.self_disp_mars = true;
		break;
        	// Jupiter
	      case '5':
		sun_options.self_disp_jupiter = true;
		break;
        	// Saturn
	      case '6':
		sun_options.self_disp_saturn = true;
		break;
        	// Uranus
	      case '7':
		sun_options.self_disp_uranus = true;
		break;
        	// Neptune
	      case '8':
		sun_options.self_disp_neptune = true;
		break;
		// what happened to poor Pluto?

	        // location
	      case 'l':
		// printf ("user specified location -l with value `%s'\n", optarg);
		abend_location(optarg);
		strcpy(sun_options.location, optarg);
		break;

		// query
	      case 'q':
		// printf ("query -q with value `%s'\n", optarg);
		strcpy(sun_options.query, optarg);
		sun_options.mode_query = true;
		break;

		// HELP
	      case '?':
	      case 'h':
		printf("orbtimes --location <geographic-coordinates> --timezone <hours east of greenwich> --daylight <offset-hours>\n");
		printf("\t--query <query-string> returns 1 if query is true, 0 otherwise\n");
                printf("\t--selftest will run exhaustive test cases\n");
		printf("\t--verbose will print diagnostic information\n");
		printf("\n");
		printf("\tExamples:\n");
		printf("\torbtimes --location \"36.048202N 114.950131W 0ft\" --timezone -8.0  --daylight 1 \n");
		printf("\tCalled with no (or partial) arguments, and it will guess at timezone, location, and daylight savings.\n");
                printf("\torbtimes --location=\"48.22N 11.54E 0ft\" -v\n");
		printf("\tCalled with --moon and/or other orbs.\n");
		printf("\torbtimes --moon --location=\"36.048621N 114.951226W\" -v\n");
		printf("\tCalled with a query\n");
		printf("\torbtimes --moon --query \"sunset>moonrise\"\n");
		printf("\tCalled on a particular date and time\n");
		printf("\torbtimes -J 2457957.0 -vv\n");
		printf("\n");
		printf("\tAll Options:\n");
		printf("\t--location (-l) %s\n", "Specify location (latitude, longitude, elevation) e.g. \"48.22N 11.54E 0ft\"");
		printf("\t--query (-q) %s\n", "Specify a query e.g. \"sunrise<sunset\"");
		printf("\t--date (-D) %s\n", "Specify the date");
		printf("\t--datejd (-J) %s\n", "Specify the Julian day");    
		printf("\t--timezone (-t) %s\n", "Specify the timezone");
		printf("\t--daylight (-d) %s\n", "Specify the daylight savings offset");
		printf("\t--twilight_angle (-e) %s\n", "Specify a twilight angle (radians)");  
		printf("\t--selftest (-s) %s\n", "Run self test mode");
		printf("\t--verbose (-v) %s\n", "Print more information");
		printf("\t--ascii_art (-A) %s\n", "Display option");
		printf("\t--moon (-M) %s\n", "Display lunar (i.e. Moon) info");
		printf("\t--mercury --venus --mars --jupiter --saturn --uranus --neptune (--orbs) %s \n", "Display each planets info (or all)");		
		printf("\t--twilight (-T) %s\n", "Display twilight info [0..3] civil,nautical,astronomical,general");
		exit(0);
		break;

	      default:
		exit(0);
	      }
    }


  /* Print any remaining command line arguments (not options). */
  if (optind < argc)
    {
      printf ("non-option ARGV-elements: ");
      while (optind < argc)
	printf ("%s ", argv[optind++]);
      putchar ('\n');
    }
  
  return 0;
}


/* latlon = vegas 
 * timezone from unix zone
 * daylight savings from unix zone
 */
void
default_options()
{
  char *old_locale, *saved_locale;
  // Geographic format is fairly flexible latitude, longitude and elevation.
  // For timekeeping, elevation is an idealization, works well for a smooth earth,
  // but not terribly effective for mountainous regions. So, for example, here in
  // Vegas the River Mountains rise in the east, and the Spring mountains are in the
  // west. Las Vegas sits in a valley which effectively makes sunrise later and sunset earlier.
  // Hence, putting an elevation value of 0ft is actually not such a bad idea. 
  // char las_vegas_location[]="36.115656N 115.172456W 1700ft"; // right click, what's here in Google maps
  //  char las_vegas_location[]="36.048202N 114.950131W 0ft"; // right click, what's here in Google maps
  char las_vegas_location[]="36.0N %fW 0ft"; // right click, what's here in Google maps    
  // Use /usr/share/zoneinfo database
  // export TZ=":America/Los_Angeles";
  if (sun_options.self_verbosity==3){
    printf("+------------------------------------------------------+\n");
    printf("timezone (zoneinfo) |%s| |%s|\n", tzname[0], tzname[1]);
  }

  // TZSET SYSTEM CALL
  tzset();
  if (sun_options.self_verbosity==3){
    printf("timezone (zoneinfo) |%s| |%s|\n", tzname[0], tzname[1]);
    printf("\t timezone=%ld seconds %lf hours west of Greenwich\n", timezone, ((double)timezone)/3600.); // timezone defined in time.h 
  }

  /* Get the name of the current locale.  */
  old_locale = setlocale (LC_ALL, NULL);

  /* Copy the name so it won’t be clobbered by setlocale. */
  saved_locale = strdup (old_locale);
  if (saved_locale == NULL)
    printf ("Out of memory");
  

  /* Now change the locale and do some stuff with it. */
  setlocale (LC_ALL, "C");
  sprintf(sun_options.location, las_vegas_location, ((double)timezone)/240.0);
  /* Restore the original locale. */
  setlocale (LC_ALL, saved_locale);
  free (saved_locale);

  sun_options.timezone = -(double)timezone/3600.0;        // negative hours west of greenwich
  sun_options.timezone_override = false;                  // automatic timezone guess
  sun_options.self_guess_daylight_savings = true;
  sun_options.daylight = 0; 
  sun_options.self_test = false;
  sun_options.self_verbose = false;  
  sun_options.self_verbosity = 0;
  sun_options.mode_query = false;
  sun_options.self_disp_twilight = false;
  sun_options.self_disp_moon = false;
  sun_options.self_disp_mercury = false;
  sun_options.self_disp_venus = false;
  sun_options.self_disp_mars = false;
  sun_options.self_disp_jupiter = false;
  sun_options.self_disp_saturn = false;
  sun_options.self_disp_uranus = false;
  sun_options.self_disp_neptune = false;    
  sun_options.mode_twilight = CIVIL_TWILIGHT;
  sun_options.self_date_is_now = false;
  sun_options.self_ascii_art = false;
}

double
guess_daylight_savings(double jd, double year, double ds)
{
  double  dst_beg = JulianDaysDOW(1,2,3,year),
    dst_end = JulianDaysDOW(1,1,11,year);
  if (dst_beg<jd and jd<dst_end) return ds;
  return 0;
}

double
guess_timezone(double longitude)
{
  double   tz;
  tz = (longitude+7.5)/15.0;
  tz = floor(tz);
  return tz;
}

void
print_options()
{
  printf("sun_options.location %s\n", sun_options.location);
  printf("sun_options.query %s\n", sun_options.query);
  printf("sun_options.date %s\n", sun_options.date);
  printf("sun_options.datejd %lf\n", sun_options.datejd);    
  printf("sun_options.timezone %lf\n", sun_options.timezone);
  printf("sun_options.timezone_override %d\n", sun_options.timezone_override);
  printf("sun_options.daylight %lf\n", sun_options.daylight);
  printf("sun_options.twilight_angle %lf\n", sun_options.twilight_angle);  
  printf("sun_options.self_test %d\n", sun_options.self_test);
  printf("sun_options.self_verbose %d\n", sun_options.self_verbose);
  printf("sun_options.self_verbosity %d\n", sun_options.self_verbosity);
  printf("sun_options.self_ascii_art %d\n", sun_options.self_ascii_art);
  printf("sun_options.self_disp_moon %d\n", sun_options.self_disp_moon);
  printf("sun_options.self_disp_twilight %d\n", sun_options.self_disp_twilight);
  printf("sun_options.self_date_is_now %d\n", sun_options.self_date_is_now);    
  printf("sun_options.self_guess_daylight_savings %d\n", sun_options.self_guess_daylight_savings);
  printf("sun_options.mode_query %d\n", sun_options.mode_query);
  printf("sun_options.mode_twilight %d\n", sun_options.mode_twilight);
}

void abend_date(char *arg)
{
  int     i;
  bool    success = true;
  char   *p = arg;
  // must match ISO8601 yyyy/mmm/dd
  while (*p){
    if (*p=='/'){ *p='-'; p++;}
    else if (*p=='\\'){ *p='-'; p++;}
    else if (*p=='-') p++;
    else if (!isalnum(*p)){ success = false; break; }
    else p++;
  }
  for (i=0;i<4;i++) success &= isdigit(arg[i]);
  success &= (arg[i]=='-');
  if (!success){
    printf("Dates must match ISO8601 yyyy-mmm-dd. Exiting...\n");
    exit(0);
  }
}

void abend_julian(char *arg)
{
  int     dec = 0;
  bool    success = true;
  char   *p = arg;
  // must match ISO8601 yyyy/mmm/dd
  while (*p){
    if (isdigit(*p)) p++;
    else if (*p=='.'){ dec++; p++; }
    else { success = false; break; }
  }
  if (!success || dec >1){
    printf("Not valid Julian day number. Exiting...\n");
    exit(0);
  }
}

/* Are fractional timezones legal? */
void abend_timezone(char *arg)
{
  int     dec = 0;
  bool    success = true;
  char   *p = arg;
  // must match ISO8601 yyyy/mmm/dd
  while (*p){
    if (isdigit(*p)) p++;
    else if (*p=='.'){ dec++; p++; }
    else if (*p=='-'){ p++; }    
    else { success = false; break; }
  }
  if (!success || dec >1){
    printf("Not valid timezone. Exiting...\n");
    exit(0);
  }
}

void abend_twilight(char *arg, int &value)
{
  char   c = *arg;
  int    success = false;
  if (c=='0'){ value=0; success=true;}
  else if (c=='1'){ value=1; success=true;}
  else if (c=='2'){ value=2; success=true;}
  else if (c=='3'){ value=3; success=true;}
  else success=false;  
  // handle civil, nautical, astronomical, general
  if (strcmp(arg,"civil")==0){ value=0; success=true; }
  else if (strncmp(arg,"naut",4)==0){ value=1; success=true; }
  else if (strncmp(arg,"astro",5)==0){ value=2; success=true; }
  else if (strncmp(arg,"gen",3)==0){ value=3; success=true; }
  else success=success;
  if (success) return;
  printf("Twilight must have values civil (0), nautical (1), astronomical (2), general (3). Exiting...\n");
  exit(0);
}

void abend_twilight_angle(char *arg)   
{
  double   value = atof(arg);
  if (fabs(value)>2*M_PI){
    printf("Invalid twilight angle. Check radians. Exiting...\n");
    exit(0);
  }
}

void abend_location(char *arg)         
{
  int     dec = 0;
  bool    success = true;
  char   *p = arg;

  while (*p){
    if (isdigit(*p)) p++;
    else if (*p=='.'){ dec++; p++; }
    else if (*p=='-'){ p++; }
    else if (*p=='N'){ p++; }
    else if (*p=='E'){ p++; }
    else if (*p=='W'){ p++; }
    else if (*p=='S'){ p++; }
    else if (*p==','){ p++; }
    else if (*p==32){ p++; }                            
    else { success = false; break; }
  }
  if (!success){
    printf("Not valid location. Exiting...\n");
    exit(0);
  }
}
