/*  Copyright (C) 2017 Nicholas C. Strauss (strauss@positive-internet.com)
    This file is part of ORBTIMES.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef __PARSEOPTIONS__
#define __PARSEOPTIONS__
#include "location_expression.h"

int parse_options (int argc, char **argv);
void default_options();
void print_options();
double guess_daylight_savings(double jd, double year, double ds);
double guess_timezone(double longitude);
void abend_date(char *arg);		   
void abend_julian(char *arg);	   
void abend_timezone(char *arg);	   
void abend_twilight(char *arg,int&value); 
void abend_twilight_angle(char *arg);   
void abend_location(char *arg);         

enum modes_twilight {
  CIVIL_TWILIGHT,
  NAUTICAL_TWILIGHT,
  ASTRONOMICAL_TWILIGHT,
  GENERAL_TWILIGHT,
};

typedef struct struct_Options
{
  char                 location[255];
  char                 query[255];
  char                 date[255];  // used on date_is_now false
  double               datejd;
  locationExpression  *latlon;
  double               timezone;   // hours west of greenwich (+)
  int                  timezone_override; // timezone set by user
  double               daylight;   // hours west of greenwich (+)
  double               twilight_angle; 
  int                  self_test;  // default is true
  int                  self_verbose; // default is false
  int                  self_verbosity;
  int                  self_ascii_art;   // default is true
  int                  self_disp_twilight;    // default is false
  int                  self_disp_moon;        // default is false
  int                  self_disp_mercury;   // default is false
  int                  self_disp_venus;   // default is false
  int                  self_disp_mars;   // default is false
  int                  self_disp_jupiter;   // default is false
  int                  self_disp_saturn;   // default is false
  int                  self_disp_uranus;   // default is false
  int                  self_disp_neptune;   // default is false
  int                  self_date_is_now; // default is true
  int                  self_guess_daylight_savings;
  int                  mode_query;
  modes_twilight       mode_twilight;   
} Options;

#endif
