/*  Copyright (C) 2017 Nicholas C. Strauss (strauss@positive-internet.com)
    This file is part of ORBTIMES.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef __STARTIME_PLANETS__
#define __STARTIME_PLANETS__

#define MAX(x,y)	((x)>(y)?(x):(y)) // generalize
#define MIN(x,y)	((x)<(y)?(x):(y))

struct planet_elements_coeff {
	extended  	mean_longitude[4];
	extended  	perihelion_longitude[4];
	extended  	eccentricity[4];
	extended  	inclination[4];
	extended  	ascending_node_longitude[4];
	extended    semi_major_axis, angular_diameter, standard_magnitude;
	void        (*perturbation)(int ip, extended t, extended ms, extended &qa,  extended &qb,
	  				extended &qc,  extended &qd,  extended &qe,  extended &qf,  extended &qg);
};

struct planet_elements {
	extended  	mean_longitude, 
				daily_longitude, 
				perihelion_longitude, 
				eccentricity, 
				inclination, 
				ascending_node_longitude,
				semi_major_axis, 
				angular_diameter, 
				standard_magnitude;
};
void compute_planet_elements(extended t,
	struct planet_elements &pe, struct planet_elements_coeff &pec);
void outer_block(
	extended	t,
	extended	&j1, 
	extended	&j2, 
	extended	&j3, 
	extended	&j4, 
	extended	&j5, 
	extended	&j6);
void jupiter_saturn_block(
	extended j2, extended j3, extended j4, 
	extended j5, extended j6,	
	extended &j7, extended &u1, extended &u2,
	extended &u3, extended &u4, extended &u5, 
	extended &u6, extended &u7, extended &u8, 
	extended &u9, extended &ua, extended &ub, 
	extended &uc, extended &ud, extended &ue, 
	extended &uf, extended &ug, extended &vh);
void uranus_neptune_block(extended t, extended j4, extended &j8, extended &j9,
	extended &vj, extended &uu, extended &uv, extended &uw);
void compute_planet(extended dj, extended ut, short planetNumber,
		extended &l0, extended &s0, extended &p0,
		extended &ep, extended &bp, extended &v0, extended &rh);
void planet_calculate_rise(short planetNumber, extended &utu, extended &utd, 
	extended dj, extended tz, extended ds,
	extended gl, extended gp, int &never_rises, int &never_sets,
	extended &au, extended &ad);
void compute_planet_position(short planetNumber, extended dj, extended tz, extended ds,
	extended gl, extended gp, extended solar_lon,
	extended &ecliptic_lat, extended &ecliptic_lon, 
	extended &ra, extended &dec, extended &value_min, extended &value_max);
void compute_planet_az_el(short planetNumber, extended dj, extended ut, extended tz, extended ds,
			  extended gl, extended gp, extended solar_lon,
			  extended &ecliptic_lat, extended &ecliptic_lon,
			  extended &ra, extended &dec, extended &value_min, extended &value_max, extended &az, extended &el);
void planet_calculate_rise_tilnow(short planetNumber, extended &utu, extended &dju, extended &utd, extended &djd,
				extended dj00, extended tz, extended ds,
				extended gl, extended gp, extended now, int now_hour, int& never_rises, int& never_sets,
				extended &au, extended &ad, int &rise_converges, int &set_converges );


#include "planetsIndex.h"

#define DEG2RAD(x) 1.745329252e-2L * (x)
#define RAD2DEG(x) 5.729577951e1L * (x)

#endif
